/* $Id: actions.c 94 2005-03-12 16:43:09Z jesus $ */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define __USE_GNU
#include <string.h>

#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#include <ncurses.h>

#include "config.h"

#include "qplay.h"
#include "data.h"
#include "edit-list.h"
#include "globals.h"

#include "parser.h"

#include "modules/mpg321.h"
#include "modules/ogg123.h"

#define FUNC(call, list, args...) /* call module function from list */ \
  list->function->call(list, ##args)

typedef struct user_command user_command;

enum argtype {
  arg_VOID,
  arg_STRING,
  arg_NUMBER
};

struct user_command {
  char *name;
  char *(*func) (char *);
  int argtype;
};

typedef struct set_var set_var;

#if 0
struct fmt_entry {
  int type;
};

struct fmt_ifclause {
  int type;
  struct fmt_entry **if_entry;
  struct fmt_entry **then_entry;
  struct fmt_entry **else_entry;
};

struct fmt_string {
  int type;
  char *str;
};

struct fmt_filler {
  int type;
  char filler;
};
#endif

struct fmt_entry **fmt;

#if 0
enum typebla {
  fmt_type_STRING,
  fmt_type_IF,
  fmt_type_FILL
};
#endif


enum vartype {
  var_type_INT,
  var_type_CHAR,
  var_type_FMTENTRY,
  var_type_STRING
};

struct set_var {
  char *name;
  int type;
  void *var;
  void *def;
};

typedef char *CMD(char *);

CMD cmd_pause;
CMD cmd_play;
CMD cmd_stop;
CMD cmd_seek;

CMD cmd_select;
CMD cmd_add;
CMD cmd_del;

CMD cmd_move;

CMD cmd_randomize;
CMD cmd_sort;

CMD cmd_repeat;
CMD cmd_shuffle;

CMD cmd_insert2queue;
CMD cmd_deletefromqueue;

CMD cmd_input;
CMD cmd_search;
CMD cmd_clear;

CMD cmd_bind;
CMD cmd_unbind;

CMD cmd_quit;

CMD cmd_set;

#ifdef DEBUG
CMD cmd_segv;
#endif /* DEBUG */

CMD cmd_unimp;

static user_command user_commands[] = {
  /* @begin */
  {"add", cmd_add, arg_STRING},
  {"del", cmd_del, arg_VOID},

  {"play", cmd_play, arg_VOID},
  {"stop", cmd_stop, arg_STRING},
  {"pause", cmd_pause, arg_STRING},
  {"seek", cmd_seek, arg_STRING},

  {"select", cmd_select, arg_STRING},
  {"move", cmd_move, arg_STRING},

  {"randomize", cmd_randomize, arg_VOID},
  {"sort", cmd_sort, arg_VOID},

  {"repeat", cmd_repeat, arg_VOID},
  {"shuffle", cmd_shuffle, arg_VOID},

  {"insert2queue", cmd_insert2queue, arg_STRING},
  {"deletefromqueue", cmd_deletefromqueue, arg_STRING},

  {"input", cmd_input, arg_STRING},
  {"search", cmd_search, arg_STRING},
  {"clear", cmd_clear, arg_VOID},

  {"bind", cmd_bind, arg_STRING},
  {"unbind", cmd_unbind, arg_STRING},

  {"set", cmd_set, arg_STRING},

#ifdef DEBUG
  {"segv", cmd_segv, arg_VOID},
#endif /* DEBUG */

  {"quit", cmd_quit, arg_VOID},
  /* @end */

  {0, 0, 0}
};

/* FIXME - only for a test: */
int foo;

static set_var set_vars[] = {
  {"format", var_type_FMTENTRY, &fmt, 0},
  {"sformat", var_type_FMTENTRY, &status_fmt, 0},
  {"foo", var_type_INT, &foo, 0}
};

char *command(char *data) {
  char *save;
  char *cmd, *rest;
  char *input;
  char *result = NULL;
  static char errbuf[512];
  user_command *uc;

  input = strdup(data);

  cmd = strtok_r(input, " ", &save);

  if (cmd == NULL) {
    free(input);
    return result;
  }

  rest = strtok_r(NULL, "\0", &save);

  PRINT_DEBUG(2, "cmd==%s; args==%s\n", cmd, rest);

  for (uc = user_commands; uc->name; uc++) {
    if (strcmp(uc->name, cmd) == 0) {
      PRINT_DEBUG(2, "call: %s %s\n", cmd, rest);
      result = uc->func(rest);
      break;
    }
  }
  if (uc->name == NULL) {
    sprintf(errbuf, "unknown command: %s", cmd);
    return errbuf;
  }

  return result;
}

QP_KEYMAP newKeymap() {
  QP_KEYMAP new;
  new = (QP_KEYMAP) malloc(sizeof(qp_keymap));
  return new;
}

void add_keybinding(QP_KEYMAP keymap, int mode, int key, char *action) {
  if (keymap->max == 0) {
    keymap->list = (QP_KEY) malloc(sizeof(qp_key));
    keymap->max = 1;
  } else if ((keymap->count + 1) > keymap->max) {
    keymap->max *= 2;
    keymap->list = (QP_KEY) realloc(keymap->list, sizeof(qp_key) * keymap->max);
  }

  keymap->list[keymap->count].keymode = 0;
  if (mode & META_MODE) {
    keymap->list[keymap->count].keymode |= META_MODE;
  }
  if (mode & CTRL_MODE) {
    if (key > 0x60) {
      key -= 0x60;
    } else if (key > 0x40) {
      key -= 0x40;
    }
  }
  keymap->list[keymap->count].keycode = key;
  keymap->list[keymap->count].action = strdup(action);
  keymap->count++;
}

static struct fmt_entry *_newEntry(enum fmt_type type) {
  struct fmt_entry *new;
  if (type == fmt_type_STRING) {
    new = (struct fmt_entry *) malloc(sizeof(struct fmt_string));
  } else if (type == fmt_type_IFCLAUSE) {
    new = (struct fmt_entry *) malloc(sizeof(struct fmt_ifclause));
    ((struct fmt_ifclause *)new)->if_entry = NULL;
    ((struct fmt_ifclause *)new)->then_entry = NULL;
    ((struct fmt_ifclause *)new)->else_entry = NULL;
  } else if (type == fmt_type_FILL) {
    new = (struct fmt_entry *) malloc(sizeof(struct fmt_filler));
    ((struct fmt_filler *)new)->rest = NULL;
  } else {
    return NULL;
  }
  new->type = type;
  return new;
}

static struct fmt_entry **_newList() {
  struct fmt_entry **new;
  new = (struct fmt_entry **) malloc(sizeof(struct fmt_entry *));
  *new = NULL;
  return new;
}

static struct fmt_entry **_add2List(struct fmt_entry **list, struct fmt_entry *new) {
  int i;

  for (i = 0; list[i] != NULL; i++);
  list = (struct fmt_entry **) realloc(list, sizeof(struct fmt_entry *) * (i + 2));
  list[i] = new;
  list[i + 1] = NULL;

  return list;
}

struct fmt_entry **str2fmt(char **fmt) {
  struct fmt_entry **new = _newList();
  char *f;
  int len = 0;

  f = *fmt;
  while (*f != 0) {
    if (*f == '%') {
      f++;
      if (*f == '{') {
        if (len > 0) {
          struct fmt_entry *newstr;
          newstr = _newEntry(fmt_type_STRING);
          ((struct fmt_string *) newstr)->str = strndup(*fmt, len);

          new = _add2List(new, newstr);
          len = 0;
        }

        struct fmt_ifclause *newif;
        struct fmt_entry **n;
        newif = (struct fmt_ifclause *) _newEntry(fmt_type_IFCLAUSE);
        f++;
        n = str2fmt(&f);
        if (*f == '?') {
          newif->if_entry = n;
          f++;
          n = str2fmt(&f);
          if (*f == ':') {
            newif->then_entry = n;
            f++;
            n = str2fmt(&f);
            newif->else_entry = n;
          } else if (*f == '}') {
            newif->then_entry = n;
          }
        } else {
          printf("parse-fehler\n");
          exit(1);
          /* fehler */
        }
        *fmt = ++f;
        new = _add2List(new, (struct fmt_entry *) newif);
        len = 0;
      } else if (*f == '>') {
        if (len > 0) {
          struct fmt_entry *newstr;
          newstr = _newEntry(fmt_type_STRING);
          ((struct fmt_string *) newstr)->str = strndup(*fmt, len);

          new = _add2List(new, newstr);
          len = 0;
        }

        struct fmt_filler *newfill;
        newfill = (struct fmt_filler *) _newEntry(fmt_type_FILL);
        f++;
        newfill->filler = *f++;
        newfill->rest = str2fmt(&f);

        new = _add2List(new, (struct fmt_entry *) newfill);
        break;
      } else {
        len += 2;
        f++;
      }
    } else if ((*f == ':') || (*f == '?') || (*f == '}')) {
      if (len > 0) {
        struct fmt_entry *newstr;
        newstr = _newEntry(fmt_type_STRING);
        ((struct fmt_string *) newstr)->str = strndup(*fmt, len);

        new = _add2List(new, newstr);
        len = 0;
      }
      break;
    } else {
      len++;
      f++;
    }
  } /* while */
  if (len > 0) {
    struct fmt_entry *newstr;
    newstr = _newEntry(fmt_type_STRING);
    ((struct fmt_string *) newstr)->str = strndup(*fmt, len);

    new = _add2List(new, newstr);
  }
  *fmt = f;

  return new;
}

#if 0
int main(int argc, char *argv[]) {
  char *fmt = ".%{%{d?bla}?daf:222}. - .%{bl?bla:jo}.";
  char *f;
  struct fmt_entry **list;
  char buf[512];

  if (argc > 1)
    fmt = argv[1];

  f = fmt;
  list = str2fmt(&f);
  parseit(list, &buf[0]);
  printf("%s\n", buf);

  return 0;
}
#endif

// extern char *default_fmt_string;

char *cmd_set(char *data) {
  char *tok, *save;
  set_var *var;

  PRINT_DEBUG(2, "set: %s\n", data);

  if (data == NULL)
    return "set required argument";

  tok = strtok_r(data, "=", &save);

  for (var = set_vars; var->name; var++) {
    if (strcmp(var->name, tok) == 0) {
      tok = strtok_r(NULL, "\0", &save);
      if (var->type == var_type_STRING) {
        if (var->var != NULL)
          free(*(char **) var->var);
        *((char **) var->var) = strdup(tok);
      } else if (var->type == var_type_FMTENTRY) {
        struct fmt_entry ***fmtvar = var->var;
        if (strcmp(tok, "default") == 0) {
          char *dummy;
          if (strcmp(var->name, "format") == 0)
            dummy = DEFAULT_FMT_STRING;
          else if (strcmp(var->name, "sformat") == 0)
            dummy = DEFAULT_STATUS_FMT_STRING;
          *fmtvar = str2fmt(&dummy);
        } else {
          *fmtvar = str2fmt(&tok);
        }
#if 0                           /* {{{ */
        char *fmtstr;
        char ch;
        char buffer[512];
        int i = 0;
        int count = 0;

        for (fmtstr = tok; *fmtstr != 0; fmtstr++) {
          if (*fmtstr == '%') {
            ch = *++fmtstr;
            if (ch == 0) {
              break;
            } else if (ch == '{') {
              struct fmt_string *dummy;
              dummy = (struct fmt_string *) malloc(sizeof(struct fmt_string));
              dummy->type = fmt_type_STRING;
              dummy->str = strndup(buffer, i);
              fmt[count] = (struct fmt_entry *) dummy;
              count++;
              i = 0;
            } else if (ch == '>') {
            } else {
              buffer[i++] = '%';
              buffer[i++] = ch;
            }
          } else if (*fmtstr == '}') {
            struct fmt_ifclause *dummy;
            dummy = (struct fmt_ifclause *) malloc(sizeof(struct fmt_ifclause));
            dummy->type = fmt_type_IF;
            dummy->if_str = strndup(buffer, i);
            fmt[count] = (struct fmt_entry *) dummy;
            count++;
            i = 0;
          } else {
            buffer[i++] = *fmtstr;
          }
        }
        struct fmt_string *dummy;

        dummy = (struct fmt_string *) malloc(sizeof(struct fmt_string));
        dummy->type = fmt_type_STRING;
        dummy->str = strndup(buffer, i);
        fmt[count] = (struct fmt_entry *) dummy;
        count++;
#endif /* }}} */
      } else if (var->type == var_type_INT) {
        *((int *) var->var) = atoi(tok);
      } else if (var->type == var_type_CHAR) {
        *((char *) var->var) = *tok;
      }
      break;
    }
  }

  return NULL;
}

/* FIXME */
CHECK_HANDLER checks[] = { mpg321_check, ogg123_check };
int modules =                   /* number of modules */
  sizeof(checks) / sizeof(CHECK_HANDLER);

char *cmd_add(char *data) {
  int j;

  PRINT_DEBUG(2, "adding: %s\n", data);
  if (data != NULL) {
    FILEENTRY new;
    new = newFileEntry(data);
    for (j = 0; j < modules; j++)
      if (checks[j] (new)) {
        addToList(filelist, new);
        break;
      }
  } else {
    return "add required argument";
  }
  return NULL;
}

char *cmd_move(char *data) {
  FILEENTRY dummy;

  if (data != NULL) {
    if (strcmp(data, "down") == 0) {
      if (filelist->selected_file < (filelist->count - 1)) {
        if (filelist->playing_file == filelist->selected_file)
          filelist->playing_file++;
        else if (filelist->playing_file == filelist->selected_file + 1)
          filelist->playing_file--;
        dummy = filelist->list[filelist->selected_file];
        filelist->list[filelist->selected_file] =
          filelist->list[filelist->selected_file + 1];
        filelist->list[filelist->selected_file + 1] = dummy;
        filelist->selected_file++;
        if (startline < filelist->selected_file - DISPLAYED_LINES)
          startline = filelist->selected_file - DISPLAYED_LINES;
      }
    } else if (strcmp(data, "up") == 0) {
      if (filelist->selected_file > 0) {
        if (filelist->playing_file == filelist->selected_file)
          filelist->playing_file--;
        else if (filelist->playing_file == filelist->selected_file - 1)
          filelist->playing_file++;
        dummy = filelist->list[filelist->selected_file];
        filelist->list[filelist->selected_file] =
          filelist->list[filelist->selected_file - 1];
        filelist->list[filelist->selected_file - 1] = dummy;
        filelist->selected_file--;
        if (filelist->selected_file < startline)
          startline = filelist->selected_file;
      }
    }
  }
  return NULL;
}

/* removes entry from playlist */
char *cmd_del(char *data) {
  int i;
  FILEENTRY old;

  PRINT_DEBUG(2, "delete: %s\n", data);
  if (filelist->selected_file >= 0) {
    old = filelist->list[filelist->selected_file];
    if (filelist->selected_file == filelist->playing_file) {
      FUNC(stop, filelist->list[filelist->playing_file]);
      filelist->playing_file = NONE;
      if (remote > 0)
        FD_CLR(remote, &fds_copy);
    }
    if (filelist->count > 0) {
      filelist->count--;
      for (i = filelist->selected_file; i < filelist->count; i++) {
        if (filelist->playing_file == i + 1)
          filelist->playing_file = i;
        filelist->list[i] = filelist->list[i + 1];
      }
      if (filelist->selected_file >= filelist->count) {
        filelist->selected_file = filelist->count - 1;
      }
    }
  }
  return NULL;
}

char *cmd_seek(char *data) {
  int seek;

  if (data == NULL)
    return "seek required argument";

  if (filelist->playing_file >= 0) {
    if (remote > 0)
      FD_CLR(remote, &fds_copy);

    if (data[0] == '+') {
      seek = atoi(data + 1);
      FUNC(skip, filelist->list[filelist->playing_file], seek);
    } else if (data[0] == '-') {
      seek = atoi(data + 1);
      FUNC(skip, filelist->list[filelist->playing_file], -seek);
    } else {
      return "wrong seek argument";
    }
    if ((remote = FUNC(getfd, filelist->list[filelist->playing_file])) > 0)
      FD_SET(remote, &fds_copy);
  }

  /* {{{
     ftime(&seeknow);
     if (seeknow.time==seeklast.time || seeknow.time-1==seeklast.time) {
     if (seeknow.millitm+1000*(seeknow.time-seeklast.time)-seeklast.millitm<100)
     seekvalue=seekvalue+1;
     } else
     seekvalue=16;
     seeklast=seeknow;
     }}} */

  return NULL;
}

char *cmd_stop(char *data) {
  PRINT_DEBUG(2, "stoping: %s\n", data);

  if (filelist->count >= 1 && filelist->playing_file >= 0) {
    FUNC(stop, filelist->list[filelist->playing_file]);
    if (remote > 0)
      FD_CLR(remote, &fds_copy);
    filelist->playing_file = NONE;
  }

  return NULL;
}

char *cmd_segv(char *data) {

  *(char *) NULL = 0;

  return NULL;
}

char *cmd_pause(char *data) {
  int new_remote;

  if (filelist->count >= 1 && filelist->playing_file >= 0)
    FUNC(pause, filelist->list[filelist->playing_file]);
  if (remote > 0)
    FD_CLR(remote, &fds_copy);
  if (filelist->count >= 1) {
    if ((new_remote = FUNC(getfd, filelist->list[filelist->playing_file])) > 0) {
      remote = new_remote;
      FD_SET(remote, &fds_copy);
    }
  }

  return NULL;
}

char *cmd_play(char *data) {
  int toplay;
  int i, newremote;
  PRINT_DEBUG(2, "playing: %s\n", data);

  if (filelist->count >= 1 && filelist->playing_file >= 0) {
    FUNC(stop, filelist->list[filelist->playing_file]);
    if (remote > 0)
      FD_CLR(remote, &fds_copy);
  }



  if (data == NULL) {
    filelist->playing_file = filelist->selected_file;
  } else if ((data[0] == '+') || (strcmp(data, "next") == 0)) {
    if (queuelist != NULL && queuelist->count > 0) {
      for (i = 0; i < filelist->count; i++) {
        if (filelist->list[i] == queuelist->list[0])
          filelist->playing_file = i;
      }
      removeFromList(queuelist, queuelist->list[0]);
    } else {
      if (data[0] == '+')
        toplay = atoi(data + 1);
      else
        toplay = +1;
      if (shuffle)
        filelist->playing_file =
          (int) (filelist->count * (rand() / (RAND_MAX + 1.0)));
      if ((filelist->playing_file + toplay) < filelist->count)
        filelist->playing_file += toplay;
      else if (repeat)
        filelist->playing_file = toplay - (filelist->count - filelist->playing_file);
      else
        filelist->playing_file = NONE;
      PRINT_DEBUG(2, "playing in number: +%d\n", toplay);
    }
  } else if ((data[0] == '-') || (strcmp(data, "prev") == 0)) {
    if (data[0] == '-')
      toplay = atoi(data + 1);
    else
      toplay = 1;
    PRINT_DEBUG(2, "playing in number: -%d\n", toplay);
    if ((filelist->playing_file - toplay) >= 0)
      filelist->playing_file -= toplay;
    else
      filelist->playing_file = NONE;
  } else {
    toplay = atoi(data);
    PRINT_DEBUG(2, "playing in number: %d\n", toplay);
    if ((toplay >= 0) && (toplay < filelist->count))
      filelist->playing_file = toplay;
    else
      filelist->playing_file = NONE;
  }

  if ((filelist->playing_file >= 0)
      && (filelist->playing_file < filelist->count)) {
    FUNC(play, filelist->list[filelist->playing_file]);
    if ((newremote = FUNC(getfd, filelist->list[filelist->playing_file])) > 0) {
      remote = newremote;
      FD_SET(remote, &fds_copy);
    }
  }

  return NULL;
}

char *cmd_search(char *data) {
  int i;
  char *pos;
  static char *search = NULL;
  char buffer[512];

  if (data != NULL) {
    if (search != NULL) {
      free(search);
    }
    search = strdup(data);
  }

  if (search != NULL) {
    for (i = (filelist->selected_file + 1); i < filelist->count; i++) {
      parseit(fmt, buffer, 512, filelist->list[i], i);
      for (pos = buffer; *pos != 0; pos++) {
        if (strncmp(pos, search, strlen(search)) == 0) {
          if (i > startline + DISPLAYED_LINES)
            startline = i - DISPLAYED_LINES;
          filelist->selected_file = i;
          return NULL;
        }
      }
      /* old {{{
      for (pos = filelist->list[i]->filename; *pos != 0; pos++) {
        if (strncmp(pos, search, strlen(search)) == 0) {
          if (i > startline + DISPLAYED_LINES) {
            startline = i - DISPLAYED_LINES;
          }
          filelist->selected_file = i;
          return NULL;
        }
      }
      }}} */
    }
    for (i = 0; i < filelist->selected_file; i++) {
      parseit(fmt, buffer, 512, filelist->list[i], i);
      for (pos = buffer; *pos != 0; pos++) {
        if (strncmp(pos, search, strlen(search)) == 0) {
          if (i < startline)
            startline = i;
          filelist->selected_file = i;
          return NULL;
        }
      }
      /* old {{{
      for (pos = filelist->list[i]->filename; *pos != 0; pos++) {
        if (strncmp(pos, search, strlen(search)) == 0) {
          if (i < startline) {
            startline = i;
          }
          filelist->selected_file = i;
          return NULL;
        }
      }
      }}} */
    }
  }

  return "not found";
}

char *cmd_randomize(char *data) {
  FILEENTRY entry;
  int i, ran;

  // filelist->status |= randomize;

  srand(getpid() * time(NULL));
  for (i = 0; i < filelist->count; i++) {
    ran = (int) (filelist->count * (rand() / (RAND_MAX + 1.0)));
    if (i == filelist->playing_file) {
      filelist->playing_file = ran;
    } else if (ran == filelist->playing_file) {
      filelist->playing_file = i;
    }
    entry = filelist->list[i];
    filelist->list[i] = filelist->list[ran];
    filelist->list[ran] = entry;
  }

  return NULL;
}

char *cmd_sort(char *data) {
  FILEENTRY entry;
  int i, j;

  /* FIXME: used the easy way (bubble-sort)
   *        but should be replaced by merge-sort */
  for (i = 0; i < (filelist->count - 1); i++) {
    for (j = 0; j < (filelist->count - 1 - i); j++) {
      if (strcmp(filelist->list[j + 1]->filename, filelist->list[j]->filename) < 0) {
        if (filelist->playing_file == (j + 1))
          filelist->playing_file = j;
        else if (filelist->playing_file == j)
          filelist->playing_file = j + 1;
        entry = filelist->list[j];
        filelist->list[j] = filelist->list[j + 1];
        filelist->list[j + 1] = entry;
      }
    }
  }

  return NULL;
}


char *cmd_repeat(char *data) {
  if (repeat)
    repeat = 0;
  else
    repeat = 1;
  return NULL;
}


char *cmd_shuffle(char *data) {
  if (shuffle)
    shuffle = 0;
  else
    shuffle = 1;
  return NULL;
}


char *cmd_select(char *data) {
  int lines;

  if (data == NULL)
    return "select required argument";

  if ((data[0] == '-') || (strcmp(data, "prev") == 0)) {
    if (data[0] == '-') {
      lines = atoi(data + 1);
    } else {
      lines = 1;
    }
    if (filelist->selected_file - lines > 0) {
      filelist->selected_file -= lines;
      if (filelist->selected_file < startline)  //scrolling up with moving the selection bar
        startline = filelist->selected_file;
      else if (filelist->selected_file > startline + DISPLAYED_LINES) { //selected_file is below view 
        if (filelist->selected_file + DISPLAYED_LINES / 2 - 1 < filelist->count)
          startline = filelist->selected_file - DISPLAYED_LINES / 2 - 1;
        else
          startline =
            filelist->selected_file - DISPLAYED_LINES + (filelist->count - 1 -
                                                         filelist->selected_file);
      }
    } else {
      filelist->selected_file = 0;
      startline = 0;
    }
  } else if ((data[0] == '+') || (strcmp(data, "next") == 0)) {
    if (data[0] == '+') {
      lines = atoi(data + 1);
    } else {
      lines = 1;
    }
    if ((filelist->selected_file + lines) < filelist->count) {
      filelist->selected_file += lines;
      if ((filelist->selected_file - startline) > DISPLAYED_LINES) {  //scrolling down with moving the selection bar
        startline = filelist->selected_file - DISPLAYED_LINES;
      } else if (startline > filelist->selected_file) { // selected_file is above view
        if (filelist->selected_file - DISPLAYED_LINES / 2 - 1 > 0) {
          startline = filelist->selected_file - DISPLAYED_LINES / 2 - 1;
        } else {
          startline = 0;
        }
      }
    } else {
      filelist->selected_file = filelist->count - 1;
      if (filelist->selected_file > DISPLAYED_LINES) {
        startline = filelist->selected_file - DISPLAYED_LINES;
      }
    }
  } else if (strcmp(data, "first") == 0) {
    filelist->selected_file = 0;
    startline = 0;
  } else if (strcmp(data, "last") == 0) {
    filelist->selected_file = filelist->count - 1;
    if (filelist->selected_file > DISPLAYED_LINES)
      startline = filelist->selected_file - DISPLAYED_LINES;
  } else if (strcmp(data, "center") == 0) {
    if (DISPLAYED_LINES < filelist->count) {
      if ((filelist->playing_file - DISPLAYED_LINES / 2 - 1) > 0) {
        startline = filelist->playing_file - DISPLAYED_LINES / 2 - 1;
      } else
        startline = 0;
      if (filelist->playing_file + DISPLAYED_LINES / 2 > filelist->count - 1) {
        startline -= DISPLAYED_LINES / 2 - filelist->count + filelist->playing_file;
      }
    }
    if (filelist->playing_file >= 0)
      filelist->selected_file = filelist->playing_file;
  } else if (strcmp(data, "down") == 0
             && startline + 1 < filelist->count - DISPLAYED_LINES) {
    startline += 1;
  } else if (strcmp(data, "up") == 0 && startline - 1 >= 0) {
    startline -= 1;
  } else if (isdigit(*data) > 0 && (lines = atoi(data)) >= 0) {
    if (lines-- < filelist->count) {
      if (lines > 0) {
        filelist->selected_file = lines;
        if ((lines - startline) > DISPLAYED_LINES)
          startline = lines - DISPLAYED_LINES;
      } else {
        filelist->selected_file = 0;
        if (filelist->selected_file < startline)
          startline = filelist->selected_file;
      }
    } else {
      filelist->selected_file = filelist->count - 1;
      if (filelist->selected_file > DISPLAYED_LINES)
        startline = filelist->selected_file - DISPLAYED_LINES;
    }
  }
  return NULL;
}

char *cmd_input(char *data) {
  if (inputmode) {              /* FIXME: this case will never arise */
    if (inputchars > 0)
      free(inputline);
    if (data != NULL) {
      inputline = strdup(data);
      inputchars = strlen(data);
    } else {
      inputchars = 0;
    }
  } else {
    if (data != NULL) {
      inputline = strdup(data);
      inputchars = strlen(data);
    } else {
      inputchars = 0;
    }
    inputmode = 1;
  }
  return data;
}

char *cmd_insert2queue(char *data) {
  filelist->list[filelist->selected_file]->inqueue = 1;
  addToList(queuelist, filelist->list[filelist->selected_file]);
  command("select next");

  /* DEBUG */
  if (debug > 1) {
    int i;
    fprintf(stderr, "\nINSERT: %i, count: %i, max: %i", filelist->selected_file,
            queuelist->count, queuelist->max);
    for (i = 0; i < queuelist->count; i++)
      fprintf(stderr, "\nfile in queue: %s", queuelist->list[i]->filename);
    fprintf(stderr, "\n");
  }
  return NULL;
}

char *cmd_deletefromqueue(char *data) {
  removeFromList(queuelist, filelist->list[filelist->selected_file]);
  command("select next");

  /* DEBUG */
  if (debug > 1) {
    int i;
    fprintf(stderr, "\nDELETE:   count: %i, max: %i", queuelist->count,
            queuelist->max);
    for (i = 0; i < queuelist->count; i++)
      fprintf(stderr, "\nfile in queue: %s", queuelist->list[i]->filename);
    fprintf(stderr, "\n");
  }
  return NULL;
}

char *cmd_osd(char *data) {
  printf("%s\n", data);
  return data;
}

char *cmd_clear(char *data) {
  rredraw = 1;
  return data;
}

struct keytab_e {
  char *keyname;
  int keycode;
};

struct keytab_e keytab[] = {
  {"<enter>", '\n'},
  {"<home>", KEY_HOME},
  {"<end>", KEY_END},
  {"<insert>", KEY_IC},
  {"<delete>", KEY_DC},
  {"<left>", KEY_LEFT},
  {"<right>", KEY_RIGHT},
  {"<up>", KEY_UP},
  {"<down>", KEY_DOWN},
  {"<pagedown>", KEY_NPAGE},
  {"<pageup>", KEY_PPAGE},
  {0, 0}
};

static int _parsekey(qp_key * key, char *keystr) {
  int control = 0;
  int i;

  key->keymode = 0;

  if (strncmp(keystr, "M-", 2) == 0) {
    key->keymode |= META_MODE;
    keystr += 2;
  }

  if (strncmp(keystr, "C-", 2) == 0) {
    control = 1;
    keystr += 2;
  }

  for (i = 0; keytab[i].keyname != NULL; i++) {
    if (strcmp(keystr, keytab[i].keyname) == 0) {
      key->keycode = keytab[i].keycode;
      break;
    }
  }

  if (keytab[i].keyname == NULL) {
    if (strlen(keystr) == 1)
      key->keycode = (int) *keystr;
    else
      return 0;
  }

  if (control) {
    if (key->keycode > 0x60) {
      key->keycode -= 0x60;
    } else if (key->keycode > 0x40) {
      key->keycode -= 0x40;
    }
  }

  return 1;
}

char *cmd_unbind(char *data) {
  qp_key key;
  int i;

  if (data == NULL) {
    return "bind required argument";
  }

  if (!_parsekey(&key, data))
    return "parse error";

  for (i = 0; i < keymap->count; i++) {
    if (keymap->list[i].keycode == key.keycode
        && keymap->list[i].keymode == key.keymode) {
      free(keymap->list[i].action);
      break;
    }
  }
  keymap->count--;
  for (; i < keymap->count; i++) {
    memcpy(keymap->list + i, keymap->list + i + 1, sizeof(qp_key));
    /* keymap->list[i].key.keycode = keymap->list[i + 1].key.keycode;
       keymap->list[i].key.keymode = keymap->list[i + 1].key.keymode;
       keymap->list[i].action = keymap->list[i + 1].action; */
  }

  return NULL;
}

char *cmd_bind(char *data) {
  qp_key key;
  char *input, *command;
  char *tok, *save;
  int i;

  if (data == NULL)
    return "bind required argument";

  input = strdup(data);

  tok = strtok_r(input, " ", &save);
  if (!_parsekey(&key, tok))
    return "error while parsing key";
  if ((tok = strtok_r(NULL, "\0", &save)) == NULL)
    return "wrong number of arguments for bind";
  command = strdup(tok);

  free(input);

  for (i = 0; i < keymap->count; i++) {
    if (keymap->list[i].keycode == key.keycode
        && keymap->list[i].keymode == key.keymode) {
      free(keymap->list[i].action);
      keymap->list[i].action = strdup(command);
      return NULL;
    }
  }

  add_keybinding(keymap, key.keymode, key.keycode, command);

  return NULL;
}

/* FIXME FIXME FIXME */
#define FIFO_FILE "/tmp/qplay-fifo"

char *cmd_quit(char *data) {

  if (filelist->count >= 1 && filelist->playing_file >= 0) {
    FUNC(stop, filelist->list[filelist->playing_file]);
  }
  endwin();
  unlink(FIFO_FILE);
  exit(EXIT_SUCCESS);

  return NULL;
}

char *cmd_unimp(char *data) {
  return data;
}

// vim600:fdm=marker:
