/* $Id: data.h 88 2005-03-07 21:24:45Z jesus $ */

#ifndef _DATA_H
#define _DATA_H

typedef struct fileentry *FILEENTRY;
typedef struct filelist *FILELIST;

/*
 * module handler
 */
typedef int (*INIT_HANDLER) (int, char **);
typedef int (*CHECK_HANDLER) (FILEENTRY);

typedef void (*PLAY_HANDLER) (FILEENTRY);
typedef void (*STOP_HANDLER) (FILEENTRY);
typedef void (*CLEANUP_HANDLER) (FILEENTRY);
typedef void (*PAUSE_HANDLER) (FILEENTRY);
typedef void (*SKIP_HANDLER) (FILEENTRY, int);

typedef void (*PARSE_HANDLER) (FILEENTRY, char *);

typedef char *(*GETNAME_HANDLER) (FILEENTRY, int);
typedef int (*TIME_HANDLER) (FILEENTRY);
typedef int (*GET_HANDLER) (FILEENTRY);

/*
 * a set of all handlers a module has
 */
typedef struct function_set {
  PLAY_HANDLER play;
  STOP_HANDLER stop;
  CLEANUP_HANDLER cleanup;
  PAUSE_HANDLER pause;
  SKIP_HANDLER skip;

  PARSE_HANDLER parse;

  GETNAME_HANDLER getname;
  TIME_HANDLER gettime;
  TIME_HANDLER gettotal;
  GET_HANDLER getfd;
  GET_HANDLER getpid;
} *FUNCTION_SET;

/*
 * file list structs
 */
struct fileentry {
  char *filename;               /* name of the file */
  FUNCTION_SET function;        /* functions of the module */
  void *module_extra_info;      /* extra information for the module */
  unsigned inqueue;             /* position in the queue */
};

struct filelist {
  int status;                   /* FIXME */
  int playing_file;             /* playing file in the list */
  int selected_file;            /* selected file in the list */

  int count;                    /* number of elements in the list */
  int max;                      /* maximum number of elements in the list */
  FILEENTRY *list;
};

enum fl_status {
  randomize = 1
};

/*
 * list of keys that bind to an actions
 */
typedef struct qp_key {
  int keycode;                  /* */
  int keymode;                  /* meta - key mode */
  char *action;                 /* action string for actions::command */
} qp_key;
typedef struct qp_key *QP_KEY;

#define META_MODE 1
#define CTRL_MODE 2

typedef struct qp_keymap {
  int count, max;               /* list counter and maximum */
  QP_KEY list;                  /* list */
} qp_keymap;
typedef struct qp_keymap *QP_KEYMAP;

/* status of list */
#define PLAYING 1
#define NONE -1

/*
 * format type for getname of the module
 */
enum str_type {
  str_type_TITLE,
  str_type_NAME,
  str_type_ARTIST,
  str_type_STATUS
};

/*
 * line format entry
 */
struct fmt_entry {
  int type;
};

struct fmt_ifclause {
  int type;
  struct fmt_entry **if_entry;
  struct fmt_entry **then_entry;
  struct fmt_entry **else_entry;
};

struct fmt_string {
  int type;
  char *str;
};

struct fmt_filler {
  int type;
  char filler;
  struct fmt_entry **rest;
};

enum fmt_type {
  fmt_type_STRING,
  fmt_type_IFCLAUSE,
  fmt_type_FILL
};


struct pathlist {
  int count, max;
  char **path;
};

#endif /* _DATA_H */
