/* $Id: mpg321.c 94 2005-03-12 16:43:09Z jesus $ */

#include <stdio.h>
#include <stdlib.h>

#define __USE_GNU
#include <string.h>

#include <sys/types.h>          /* socketpair & fork & waitpid & open */
#include <sys/socket.h>         /* socketpair */
#include <unistd.h>             /* fork */
#include <signal.h>             /* kill */

#include <sys/wait.h>           /* waitpid */

#include <time.h>               /* time */

#include <unistd.h>             /* getopt_long */

#define _GNU_SOURCE
#include <getopt.h>

#include <sys/stat.h>           /* open */

#define __USE_LARGEFILE64
#include <fcntl.h>              /* open */

#include "data.h"
#include "config.h"

extern int playing;             /* position of the playing song */

/* predefinitions of the functions to export */
void mpg321_parse_line(FILEENTRY, char *);
void mpg321_play(FILEENTRY);
void mpg321_stop(FILEENTRY);
void mpg321_cleanup(FILEENTRY);
void mpg321_pause(FILEENTRY);
void mpg321_skip(FILEENTRY, int);

int mpg321_get_time(FILEENTRY);
int mpg321_get_total(FILEENTRY);
int mpg321_get_fd(FILEENTRY);
int mpg321_get_pid(FILEENTRY);

char *mpg321_getstr(FILEENTRY, int);
int mpg321_init(int, char **);

void mpg321_nothing();

/* module functions (exported) */
struct function_set MPG321_FUNC_SET = {
  .play = mpg321_play,
  .parse = mpg321_parse_line,
  .stop = mpg321_stop,
  .cleanup = mpg321_cleanup,
  .pause = mpg321_pause,
  .gettime = mpg321_get_time,
  .gettotal = mpg321_get_total,
  .skip = mpg321_skip,

  .getfd = mpg321_get_fd,
  .getpid = mpg321_get_pid,
  .getname = mpg321_getstr,
};

#define PAUSED 1
#define BROKEN 2
#define HTTP 4
#define FTP 8

/* informations of a song entry */
struct extra_info_struct {
  /*
     pid_t pid;
     int fd;
   */
  int status;
  int time, total;
  int frames, aframe;

  /* id3-tag info */
  char *title, *artist, *album;
  char *comment, *year;
  int track, genre;
};

/* 005359A0   54 41 47 54 69 74 6C 65  20 20 20 20 20 20 20 20  TAGTitle
 * 005359B0   20 20 20 20 20 20 20 20  20 20 20 20 20 20 20 20
 * 005359C0   20 41 75 74 68 6F 72 20  20 20 20 20 20 20 20 20   Author
 * 005359D0   20 20 20 20 20 20 20 20  20 20 20 20 20 20 20 62                 b
 * 005359E0   6C 61 20 20 20 20 20 20  20 20 20 20 20 20 20 20  la
 * 005359F0   20 20 20 20 20 20 20 20  20 20 20 20 20 31 39 39               199
 * 00535A00   30 62 6C 61 20 20 20 20  20 20 20 20 20 20 20 20  0bla
 * 00535A10   20 20 20 20 20 20 20 20  20 20 20 20 20 00 34 4E               .4N
 */

/* File: /home/jesus/mucke/Ace Of Base # Que Sera.mp3
 * Title:   Que Sera                       Track: 
 * Artist:  Ace of Base
 * Album:   The Bridge                     Year:  
 * Comment: Track 13                       Genre: Soundtrack [24]
 */

static int fd;                  /* file descriptor to receive data of the playing program */
static pid_t pid;               /* process id of the playing program */

// static struct extra_info_struct extra_info;

/*
 * do nothing function
 */
void mpg321_nothing() {
  /* noop */
}

/* where to find the player program */
static char *program = NULL;

/*
 * initials the module
 */
int mpg321_init(int argc, char *argv[]) {
  char *path;
  char *tok, *save;
  char file[512];
  /*
     char *dpoint;
     char *pathenv;
     // char path[128];

     pathenv = getenv("PATH");
     dpoint = pathenv;

     while (dpoint != NULL) {
     dpoint = index(dpoint, ':');
     printf("%s\n", dpoint++);
     }
     sleep(2);
   */
  path = strdup(getenv("PATH"));

  for (tok = strtok_r(path, ":", &save); tok != NULL;
       tok = strtok_r(NULL, ":", &save)) {
    sprintf(file, "%s/%s", tok, "mpg123");
    if (access(file, X_OK) == 0) {
      program = strdup(file);
      break;
    }
  }
  // program = strdup("mpg123");

  return 1;
}

/*
 * pauses a song
 */
void mpg321_pause(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;
  pid_t opid;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->status & PAUSED) {
    fentry->function->play(fentry);
  } else {
    extra_info->status |= PAUSED;
    if (fd > 0) {
      close(fd);
      fd = -1;
    }
    if (pid > 0) {              /* if there is a player .. */
      kill(pid, SIGINT);        /* .. stop it */
      opid = waitpid(pid, NULL, WNOHANG); /* .. and wait for it */
      if (pid != opid && opid != -1) {  /* if a problem .. */
        waitpid(pid, NULL, 0);  /* .. wait again */
      }
      pid = 0;                  /* marks that no more process is there */
    }
  }
}

/*
 * FIXME: no more in use
 */
static int _start_pos = 0;
static int cut_path = 0;

/*
 * compares two strings and returns the position of the difference
 */
static int _compare(const char *str1, const char *str2) {
  int i;

  for (i = 0; (i < strlen(str1)) && (i < strlen(str2)); i++) {
    if (str1[i] != str2[i])
      break;
  }
  return i;
}

/*
 * cut spaces at the end of a string
 */
static char *_trim(char *str, int len) {
  int i;
  char *ret;

  for (i = len; (str[i - 1] == ' ') && (i > 0); i--)  /* noop */
    ;

  if (i == 0)
    ret = NULL;
  else
    ret = strndup(str, i);

  return ret;
}

/*
 * checks responsibility for a file entry
 */
int mpg321_check(FILEENTRY fentry) {
  static char save[512];
  char *extension = fentry->filename;
  struct extra_info_struct *extra_info;
  int new_start;

  if (strlen(extension) > 3) {

#if 0                           /* {{{ old stuff, is to be removed */
#endif /* }}} */
    if (cut_path) {             /* little nice hack :) */
      if (_start_pos == 0) {
        strncpy(save, extension, 511);
        _start_pos = strlen(extension);
      } else {
        new_start = _compare(save, extension);
        if (new_start < _start_pos) {
          strncpy(save, extension, new_start);
          _start_pos = new_start;
        }
      }
    }

    /*
     * FIXME: there should be three modes for reading id3-TAG's
     *        first:    read while loading file into the list (here)
     *        secound:  read when printing the file entry
     *        third:    read during playing the song
     */
    extension += strlen(extension) - 4;
    /* FIXME: don't check only the extension */
    if (strcmp(extension, ".mp3") == 0) {
      if (strncmp(fentry->filename, "http://", 7) == 0) {
        extra_info = malloc(sizeof(struct extra_info_struct));
        memset(extra_info, 0, sizeof(struct extra_info_struct));
        fentry->function = &MPG321_FUNC_SET;  /* sets modul functions to the file entry */
        fentry->module_extra_info = (void *) extra_info;  /* also sets extra informations for each entry */
        extra_info->status |= HTTP;
        return 1;
      } else if (strncmp(fentry->filename, "ftp://", 6) == 0) {
        extra_info = malloc(sizeof(struct extra_info_struct));
        memset(extra_info, 0, sizeof(struct extra_info_struct));
        fentry->function = &MPG321_FUNC_SET;  /* sets modul functions to the file entry */
        fentry->module_extra_info = (void *) extra_info;  /* also sets extra informations for each entry */
        extra_info->status |= FTP;
        return 1;
      } else {
        char buf[30];
        int track = 0, genre = 0;
        int infile;

        if ((infile = open(fentry->filename, O_RDONLY | O_LARGEFILE)) > 0) {
          extra_info = malloc(sizeof(struct extra_info_struct));
          memset(extra_info, 0, sizeof(struct extra_info_struct));

          lseek(infile, -0x80, SEEK_END);
          read(infile, buf, 3); /* reads TAG-id */
          if (strncmp(buf, "TAG", 3) == 0) {  /* if there is a TAG */
            read(infile, buf, 30);  /* .. read title */
            extra_info->title = _trim(buf, 30);
            read(infile, buf, 30);  /* .. read artist */
            extra_info->artist = _trim(buf, 30);
            read(infile, buf, 30);  /* .. read album */
            extra_info->album = _trim(buf, 30);
            read(infile, buf, 4); /* .. read year */
            extra_info->year = _trim(buf, 4);
            read(infile, buf, 30);  /* .. read comment */
            extra_info->comment = _trim(buf, 30);
            read(infile, &track, 1);  /* .. read track number */
            read(infile, &genre, 1);  /* .. read genre - FIXME: create a list of strings for it */
          }
          close(infile);

          fentry->function = &MPG321_FUNC_SET;  /* sets modul functions to the file entry */
          fentry->module_extra_info = (void *) extra_info;  /* also sets extra informations for each entry */
          return 1;
        }
      }
    }
  }
  return 0;
}

/*
 * FIXME: seizes the following functions together to one
 *        - mpg321_get_fd
 *        - mpg321_get_pid
 *        - mpg321_get_time
 *        - mpg321_get_total
 *        maybe likewise the getstr-function
 */

int mpg321_get_fd(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return fd;
}

int mpg321_get_pid(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return pid;
}

int mpg321_get_time(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->time;
}

int mpg321_get_total(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->total;
}

#define CUT(line, chr) \
  while ((*line != 0) && (*line == chr)) \
    line++;

#define UPTO(line, chr) \
  line; \
  while ((*line != 0) && (*line != chr)) \
    line++; \
  *line++ = 0;

/* (local)
 * parses the time part
 * FIXME: not for all mpg123-versions suitable
 */
static int _mpg321_parse_time(char *time_str) {
  char *minutes, *seconds /*, *useconds */ ;
  int min, sec /*, usec */ ;

  minutes = time_str;
  while ((*time_str >= '0') && (*time_str <= '9'))
    time_str++;
  *time_str++ = 0;
  min = atoi(minutes);

  seconds = time_str;
  while ((*time_str >= '0') && (*time_str <= '9'))
    time_str++;
  *time_str++ = 0;
  sec = atoi(seconds);

  return ((min * 60) + sec);
}

/* (local) used, in order to display the time of broken files */
static int starttime;

/*
 * parses the received line of the player
 */
void mpg321_parse_line(FILEENTRY fentry, char *line) {
  char *tok, *save;
  struct extra_info_struct *extra_info;
  char *timeline, *frameline;
  int old_aframe;

  // Frame#    54 [ 8490], Time: 00:01.41 [03:41.77],
  while (*line != 0) {
    /* FIXME: works not for all mpg123-versions */
    if (strncmp(line, "Frame#", 5) == 0) {
      extra_info = (struct extra_info_struct *) fentry->module_extra_info;

      frameline = strtok_r(line, ",", &save);
      timeline = strtok_r(NULL, ",", &save);

      frameline += 6;
      CUT(frameline, ' ');
      tok = UPTO(frameline, ' ');
      old_aframe = extra_info->aframe;
      extra_info->aframe = atoi(tok);

      CUT(frameline, ' ');
      CUT(frameline, '[');
      CUT(frameline, ' ');
      tok = UPTO(frameline, ']');
      if (extra_info->frames == 0)
        extra_info->frames = atoi(tok);

      if (old_aframe > (extra_info->aframe + 200)) {
        extra_info->status |= BROKEN;
        extra_info->frames = 0;
        // printf("fehler %d %d\n", old_aframe, extra_info->aframe);
      }

      if (timeline != NULL) {
        CUT(timeline, ' ');
        if (strncmp(timeline, "Time:", 5) == 0) {
          timeline += 5;
          CUT(timeline, ' ');
          tok = UPTO(timeline, ' ');
/* FIXME REMOVEME         if (extra_info->status & BROKEN) {
            extra_info->time = time(NULL) - starttime;
          } else { */
          extra_info->time = _mpg321_parse_time(tok);
/*          } */

          CUT(timeline, ' ');
          CUT(timeline, '[');
          CUT(timeline, ' ');
          tok = UPTO(timeline, ']');
          if (extra_info->total == 0)
            extra_info->total = _mpg321_parse_time(tok);
        }
      }
    }
    line++;
  }
}

/*
 * returns file information
 */
char *mpg321_getstr(FILEENTRY fentry, int type) {
  struct extra_info_struct *extra_info;
  char *tok, *save, *last;
  char *filename;
  static char buffer[512];
  char *ch = buffer;            /* for status only */

  extra_info = ((struct extra_info_struct *) fentry->module_extra_info);

  /* FIXME: here we can read information, if this did not happen yet. */
  switch (type) {
  case str_type_TITLE:
    return extra_info->title;
    break;
  case str_type_NAME:
    filename = strdup(fentry->filename);
    last = tok = strtok_r(filename, "/", &save);
    while ((tok = strtok_r(NULL, "/", &save)) != NULL)
      last = tok;

    tok = last;
    tok += strlen(last) - 4;
    *tok = 0;

    strcpy(buffer, last);
    free(filename);
    return buffer;
    break;
  case str_type_ARTIST:
    return extra_info->artist;
    break;
  case str_type_STATUS:
    if (extra_info->status & HTTP)
      *ch++ = 'h';
    if (extra_info->status & FTP)
      *ch++ = 'f';
    if (extra_info->status & BROKEN)
      *ch++ = 'b';
    if (extra_info->status & PAUSED)
      *ch++ = 'p';
    *ch = 0;
    return buffer;
    break;
  }

  /*
     static char dummy[512];
     dummy[0] = 0;
     for (i = 0; i < strlen(format); i++) {
     if (format[i] == '$') {
     if (++i >= strlen(format))
     break;
     switch (format[i]) {
     case 't':
     if (extra_info->title != NULL)
     strcat(dummy, extra_info->title);
     break;
     case 'a':
     if (extra_info->artist != NULL)
     strcat(dummy, extra_info->artist);
     break;
     case 'f':
     if (fentry->filename != NULL)
     strcat(dummy, fentry->filename);
     break;
     default:
     break;
     }
     } else {
     strncat(dummy, &format[i], 1);
     }
     }
   */

/*
  if (extra_info->status & BROKEN) {
    sprintf(dummy, "%s  (damaged)", fentry->filename + _start_pos);
  } else {
    if (extra_info->title != NULL) {
      sprintf(dummy, "%s # %s", extra_info->artist, extra_info->title);
    } else {
      sprintf(dummy, "%s", fentry->filename + _start_pos);
    }
  }
*/

  return NULL;
}

/* FIXME: don't execute mpg-player with shell (not in use) */
#define SHELL "/bin/sh"

/*
 * plays a song
 */
void mpg321_play(FILEENTRY fentry) {
  char *program = "mpg321";
  char webname[1024];
  int fdpipe[2];
  pid_t child_pid;
  struct extra_info_struct *extra_info;
  char *argv[10];               /* FIXME */
  int argc = 0;
  char buf[32];

  extra_info = ((struct extra_info_struct *) fentry->module_extra_info);

  extra_info->status &= ~PAUSED;

  if (extra_info->status & FTP) {
    pid_t ncftp_pid;
    int com[2];

    pipe(com);

    if ((ncftp_pid = fork()) == 0) {     /* first child */
      close(com[0]);
      dup2(com[1], 1);
      dup2(com[1], 2);
      argc = 0;
      argv[argc++] = "ncftpget";
      argv[argc++] = "-c";
      argv[argc++] = fentry->filename;
      argv[argc++] = NULL;
      execvp("ncftpget", argv);
      exit(1);
    } else {                            /* mother */
      close(com[1]);
      pipe(fdpipe);
      if ((child_pid = fork()) == 0) {  /* second child */
        dup2(com[0], 0);
        close(fdpipe[0]);
        dup2(fdpipe[1], 1);
        dup2(fdpipe[1], 2);
        argc = 0;
        argv[argc++] = program;
        argv[argc++] = "-v";
        argv[argc++] = "-";
        argv[argc++] = NULL;
        execvp(program, argv);
        exit(1);
      } else {                          /* mother */
        close(com[0]);
        close(fdpipe[1]);

        pid = child_pid;
        fd = fdpipe[0];
        return;
      }
    }
  }

  argc = 0;
  argv[argc++] = program;
  argv[argc++] = "--skip-printing-frames=16";
#ifdef TESTMODE
  argv[argc++] = "-t";
#endif
  if (extra_info->aframe > 0) {
    sprintf(buf, "-k%d", extra_info->aframe);
    argv[argc++] = buf;
  }
  argv[argc++] = "-v";
  if (extra_info->status & HTTP) {
    char *ch;
    char *dummy = webname;
    for (ch = fentry->filename; *ch != 0; ch++) {
      if (*ch == '#') {
        dummy += sprintf(dummy, "%%23");
      } else {
        *dummy++ = *ch;
      }
    }
    *dummy = 0;
    argv[argc++] = webname;
  } else {
    argv[argc++] = fentry->filename;
  }
  argv[argc++] = NULL;

  /* Prepares socket to receive output of the player */
  /* FIXME: we don't need a bi-directional socket,
   * pipe(2) will completely be sufficient. */
  socketpair(AF_UNIX, SOCK_STREAM, 0, fdpipe);

  /* A new child, in order to start the player */
  if ((child_pid = fork()) == 0) {  /* child */
    close(0);
    close(fdpipe[1]);
    // dup2(fdpipe[0], 0);
    dup2(fdpipe[0], 1);
    dup2(fdpipe[0], 2);
    /* Creates a new session. Don't know why.. :)
     * But it seems to affect the child with segfault */
    setsid();
    execvp(program, argv);
    // execlp(program, program, "-v", fentry->filename, NULL);
    exit(EXIT_FAILURE);
  } else if (child_pid > 0) {   /* mother */
    /* FIXME: this is only used for broken player */
    starttime = time(NULL);

    close(fdpipe[0]);
    pid = child_pid;
    fd = fdpipe[1];
    // return fdpipe[1];
  } else if (child_pid < 0) {   /* error */
  }
  // return -1;
}

/*
 * cleans players-information of a song
 */
void mpg321_cleanup(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;
  pid_t old_pid;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;

  if (pid > 0) {
    if (fd > 0) {
      close(fd);
      fd = -1;
    }
    old_pid = waitpid(pid, NULL, WNOHANG);
    if (pid != old_pid && old_pid != -1) {
      waitpid(pid, NULL, 0);
    }
    pid = 0;
    extra_info->aframe = 0;
    extra_info->time = 0;
    extra_info->status &= ~PAUSED;
  }
}

/*
 * stops playing a song
 */
void mpg321_stop(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (pid > 0) {
    kill(pid, SIGINT);
    mpg321_cleanup(fentry);
  }
}

/*
 * jumps within a song
 */
void mpg321_skip(FILEENTRY fentry, int skip) {
  struct extra_info_struct *extra_info;
  int nframe, ntime;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->status & FTP)
    return;
  if (pid > 0) {
    /* don't jump if the file were marked as broken */
    // if ((extra_info->status & BROKEN) == 0) {
    /* FIXME: we can jump only by frame */
    nframe = extra_info->aframe + (skip * 200);
    ntime = extra_info->time + (skip * 5);
    if (nframe < 0) {
      nframe = 0;
      ntime = 0;
    }
    if (nframe >= extra_info->frames && extra_info->frames > 0) {
      nframe = extra_info->frames - 40;
      ntime = extra_info->total - 1;
    }
    mpg321_stop(fentry);
    extra_info->aframe = nframe;
    extra_info->time = ntime;   /* FIXME: little hack */
    mpg321_play(fentry);
    // }
  }
}
// vim600:fdm=marker:
