/* $Id: ogg123.c 94 2005-03-12 16:43:09Z jesus $ */

#include <stdio.h>
#include <stdlib.h>

#define __USE_GNU
#include <string.h>

#include <sys/types.h>          /* socketpair & fork & waitpid & open */
#include <sys/socket.h>         /* socketpair */
#include <unistd.h>             /* fork */
#include <signal.h>             /* kill */
#include <sys/wait.h>           /* waitpid */

#include <sys/stat.h>           /* open */

#define __USE_LARGEFILE64
#include <fcntl.h>              /* open */

#include "data.h"
#include "config.h"

extern int playing;

void ogg123_parse_line(FILEENTRY, char *);
void ogg123_play(FILEENTRY);
void ogg123_stop(FILEENTRY);
void ogg123_cleanup(FILEENTRY);
void ogg123_pause(FILEENTRY);
int ogg123_get_time(FILEENTRY);
int ogg123_get_total(FILEENTRY);
void ogg123_skip(FILEENTRY, int);

int ogg123_get_fd(FILEENTRY);
int ogg123_get_pid(FILEENTRY);
char *ogg123_getstr(FILEENTRY, int);

void ogg123_nothing();

struct function_set OGG123_FUNC_SET = {
  .play = ogg123_play,
  .parse = ogg123_parse_line,
  .stop = ogg123_stop,
  .cleanup = ogg123_cleanup,
  .pause = ogg123_pause,
  .gettime = ogg123_get_time,
  .gettotal = ogg123_get_total,
  .skip = ogg123_skip,

  .getfd = ogg123_get_fd,
  .getpid = ogg123_get_pid,
  .getname = ogg123_getstr,
};

#define PAUSED 1
#define HTTP 2

struct extra_info_struct {
  pid_t pid;
  int fd;
  int status;
  int time, total;

  char *title, *artist, *album;
  char *comment, *year;
  char *track, *genre;
};

/* 00000000   4F 67 67 53 00 02 00 00  00 00 00 00 00 00 23 B3  OggS..........#.
 * 00000010   6D 08 00 00 00 00 C5 94  1E A4 01 1E 01 76 6F 72  m............vor
 * 00000020   62 69 73 00 00 00 00 02  44 AC 00 00 FF FF FF FF  bis.....D.......
 * 00000030   00 F4 01 00 FF FF FF FF  B8 01 4F 67 67 53 00 00  ..........OggS..
 * 00000040   00 00 00 00 00 00 00 00  23 B3 6D 08 01 00 00 00  ........#.m.....
 * 00000050   17 2A 1A FB 11 FF 25 FF  FF FF FF FF FF FF FF FF  .*....%.........
 * 00000060   FF FF FF FF FF FF 03 76  6F 72 62 69 73 20 00 00  .......vorbis ..
 * 00000070   00 58 69 70 68 6F 70 68  6F 72 75 73 20 6C 69 62  .Xiphophorus lib
 * 00000080   56 6F 72 62 69 73 20 49  20 32 30 30 31 30 32 32  Vorbis I 2001022
 * 00000090   35 06 00 00 00 79 00 00  00 54 68 69 73 20 73 6F  5....y...This so
 * 000000A0   6E 67 20 69 73 20 70 61  72 74 20 6F 66 20 74 68  ng is part of th
 * 000000B0   65 20 4F 70 65 6E 20 4D  75 73 69 63 20 50 72 6F  e Open Music Pro
 * 000000C0   6A 65 63 74 2E 20 49 74  20 69 73 20 63 6F 76 65  ject. It is cove
 * 000000D0   72 64 20 62 79 20 74 68  65 20 67 72 65 65 6E 20  rd by the green
 * 000000E0   4F 70 65 6E 4D 75 73 69  63 20 4C 69 63 65 6E 63  OpenMusic Licenc
 * 000000F0   65 20 28 68 74 74 70 3A  2F 2F 6F 70 65 6E 6D 75  e (http://openmu
 * 00000100   73 69 63 2E 6C 69 6E 75  78 74 61 67 2E 6F 72 67  sic.linuxtag.org
 * 00000110   2F 29 11 00 00 00 74 69  74 6C 65 3D 4F 70 65 6E  /)....title=Open
 * 00000120   20 53 6F 75 72 63 65 16  00 00 00 61 72 74 69 73   Source....artis
 */

// static struct extra_info_struct extra_info;

void ogg123_nothing() {
  /* noop */
}

int ogg123_init(int argc, char *argv[]) {
  return 1;
}

void ogg123_pause(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;
  // pid_t pid;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->status & HTTP)  /* ogg123 can't jump in http-links */
    return;
  if (extra_info->status & PAUSED) {
    fentry->function->play(fentry);
  } else {
    extra_info->status |= PAUSED;
    if (extra_info->fd > 0) {
      close(extra_info->fd);
      extra_info->fd = -1;
    }
    if (extra_info->pid > 0) {
      kill(extra_info->pid, SIGINT);
      // pid = waitpid(extra_info->pid, NULL, WNOHANG);
      // if (pid != extra_info->pid && pid != -1)
      waitpid(extra_info->pid, NULL, 0);
      extra_info->pid = 0;
    }
  }
}

char *ogg123_getstr(FILEENTRY fentry, int type) {
  struct extra_info_struct *extra_info;
  char *tok, *save, *last;
  char *filename;
  static char buffer[512];
  char *ch = buffer;            /* for status only */

  extra_info = ((struct extra_info_struct *) fentry->module_extra_info);

  switch (type) {
  case str_type_TITLE:
    return extra_info->title;
    break;
  case str_type_NAME:
    filename = strdup(fentry->filename);
    last = tok = strtok_r(filename, "/", &save);
    while ((tok = strtok_r(NULL, "/", &save)) != NULL)
      last = tok;

    tok = last;
    tok += strlen(last) - 4;
    *tok = 0;

    strcpy(buffer, last);
    free(filename);
    return buffer;
    break;
  case str_type_ARTIST:
    return extra_info->artist;
    break;
  case str_type_STATUS:
    if (extra_info->status & HTTP)
      *ch++ = 'h';
    if (extra_info->status & PAUSED)
      *ch++ = 'p';
    *ch = 0;
    return buffer;
    break;
  }

  return NULL;
}

int ogg123_check(FILEENTRY fentry) {
  char *extension = fentry->filename;
  struct extra_info_struct *extra_info;

  if (strlen(extension) > 3) {
    extension += strlen(extension) - 4;
    if (strcmp(extension, ".ogg") == 0) {
      if (strncmp(fentry->filename, "http://", 7) == 0) {
        extra_info = malloc(sizeof(struct extra_info_struct));
        memset(extra_info, 0, sizeof(struct extra_info_struct));
        fentry->function = &OGG123_FUNC_SET;  /* sets modul functions to the file entry */
        fentry->module_extra_info = (void *) extra_info;  /* also sets extra informations for each entry */
        extra_info->status |= HTTP;
        return 1;
      } else {
        int infile;
        if ((infile = open(fentry->filename, O_RDONLY | O_LARGEFILE)) > 0) {
          char buf[512];
          int count = 0, slen;

          extra_info = malloc(sizeof(struct extra_info_struct));
          memset(extra_info, 0, sizeof(struct extra_info_struct));

          lseek(infile, 0, SEEK_SET);
          read(infile, buf, 4);
          if (strncmp(buf, "OggS", 4) == 0) {
            int i;
            lseek(infile, 145, SEEK_SET);
            read(infile, &count, 4);
            for (i = 0; i < count && i < 5; i++) {
              read(infile, &slen, 4);
              if (slen > 512)
                slen = 512;
              read(infile, buf, slen);
              if (strncmp(buf, "title=", 6) == 0) {
                extra_info->title = strndup(buf + 6, slen - 6);
              } else if (strncmp(buf, "artist=", 7) == 0) {
                extra_info->artist = strndup(buf + 7, slen - 7);
              } else if (strncmp(buf, "album=", 6) == 0) {
                extra_info->album = strndup(buf + 6, slen - 6);
              } else if (strncmp(buf, "tracknumber=", 12) == 0) {
                extra_info->track = strndup(buf + 12, slen - 12);
              }
            }
          }
          close(infile);

          fentry->function = &OGG123_FUNC_SET;
          fentry->module_extra_info = (void *) extra_info;
          return 1;
        }
      }
    }
  }
  return 0;
}

int ogg123_get_fd(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->fd;
}

int ogg123_get_pid(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->pid;
}

int ogg123_get_time(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->time;
}

int ogg123_get_total(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  return extra_info->total;
}

#define CUT(line, chr) \
  while ((*line != 0) && (*line == chr)) \
    line++;

#define UPTO(line, chr) \
  line; \
  while ((*line != 0) && (*line != chr)) \
    line++; \
  *line++ = 0;

int _ogg123_parse_time(char *time_str) {
  char *minutes, *seconds /*, *useconds */ ;
  int min, sec /*, usec */ ;

  minutes = time_str;
  while ((*time_str >= '0') && (*time_str <= '9'))
    time_str++;
  *time_str++ = 0;
  min = atoi(minutes);

  seconds = time_str;
  while ((*time_str >= '0') && (*time_str <= '9'))
    time_str++;
  *time_str++ = 0;
  sec = atoi(seconds);

  return ((min * 60) + sec);
}

void ogg123_parse_line(FILEENTRY fentry, char *line) {
  char *tok;
  struct extra_info_struct *extra_info;
  char *timeline;

  // Time: 00:07.83 [03:45.13] of 03:52.96  (127.3 kbps)  Output Buffer  96.9%
  while (*line != 0) {
    if (strncmp(line, "Time:", 5) == 0) {
      extra_info = (struct extra_info_struct *) fentry->module_extra_info;
      timeline = line;

      timeline += 5;
      CUT(timeline, ' ');
      tok = UPTO(timeline, ' ');
      extra_info->time = _ogg123_parse_time(tok);

      /* FIXME should read the whole ("of 03:52.96") time from line
       * instead of "[03:45.13]" */
      CUT(timeline, ' ');
      CUT(timeline, '[');
      CUT(timeline, ' ');
      tok = UPTO(timeline, ']');
      if (extra_info->total == 0)
        extra_info->total = _ogg123_parse_time(tok);
    }
    line++;
  }
}

#define SHELL "/bin/sh"

void ogg123_play(FILEENTRY fentry) {
  char ogg123call[512];
  int fdpipe[2];
  pid_t pid;
  struct extra_info_struct *extra_info;
  char *argv[16];               /* FIXME */
  int argc = 0;
  char buf[32];

  extra_info = ((struct extra_info_struct *) fentry->module_extra_info);

#define OGG123 "ogg123"

  extra_info->status &= ~PAUSED;
  if (extra_info->time > 0) {
    sprintf(ogg123call, OGG123 " -v -k %d '%s'", extra_info->time, fentry->filename);
  } else {
    sprintf(ogg123call, OGG123 " -v '%s'", fentry->filename);
  }

  argc = 0;
  argv[argc++] = OGG123;
#ifdef TESTMODE
  argv[argc++] = "-d";
  argv[argc++] = "raw";
  argv[argc++] = "-f";
  argv[argc++] = "/dev/null";
#endif
  if (extra_info->time > 0) {
    sprintf(buf, "-k%d", extra_info->time);
    argv[argc++] = buf;
  }
  argv[argc++] = "-v";
  argv[argc++] = fentry->filename;
  argv[argc++] = NULL;

  socketpair(AF_UNIX, SOCK_STREAM, 0, fdpipe);

  if ((pid = fork()) == 0) {    /* child */
    close(fdpipe[1]);
    dup2(fdpipe[0], 0);
    dup2(fdpipe[0], 1);
    dup2(fdpipe[0], 2);
    setsid();
    execvp(OGG123, argv);
    // execlp(SHELL, SHELL, "-c", ogg123call, NULL);
    exit(EXIT_FAILURE);
  } else if (pid > 0) {         /* mother */
    close(fdpipe[0]);
    extra_info->pid = pid;
    extra_info->fd = fdpipe[1];
  }
}

void ogg123_cleanup(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;
  /* pid_t pid; */

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->pid > 0) {
    if (extra_info->fd > 0) {
      close(extra_info->fd);
      extra_info->fd = -1;
    }
    // pid = waitpid(extra_info->pid, NULL, WNOHANG);
    // if (pid != extra_info->pid && pid != -1)
    waitpid(extra_info->pid, NULL, 0);
    extra_info->pid = 0;
    /* extra_info->aframe = 0; */
    extra_info->time = 0;
    extra_info->status = 0;
  }
}

void ogg123_stop(FILEENTRY fentry) {
  struct extra_info_struct *extra_info;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->pid > 0) {
    /* close(extra_info->fd); */
    kill(extra_info->pid, SIGINT);
    ogg123_cleanup(fentry);
  }
}

void ogg123_skip(FILEENTRY fentry, int skip) {
  struct extra_info_struct *extra_info;
  int ntime;

  extra_info = (struct extra_info_struct *) fentry->module_extra_info;
  if (extra_info->status & HTTP)  /* ogg123 can't jump in http-files */
    return;
  if (extra_info->pid > 0) {
    ntime = extra_info->time + (skip * 5);
    if (ntime < 0) {
      ntime = 0;
    }
    if (ntime >= extra_info->total) {
      ntime = extra_info->total - 1;
    }
    ogg123_stop(fentry);
    extra_info->time = ntime;
    ogg123_play(fentry);
  }
}
