/* $Id: parser.c 94 2005-03-12 16:43:09Z jesus $ */

#include <stdio.h>
#include <string.h>

#include "edit-list.h"
#include "globals.h"
#include "data.h"

/*
 * parses one option
 */
int parse_op(char op, char *prefix, char *buffer, int max, FILEENTRY file, int index) {
  int len = 0;
  char *str;
  int time, total;
  char fmt[512], dummy[512];

  *buffer = 0;
  if (op == 't') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((str = FUNC(getname, file, str_type_TITLE)) != NULL)
      len = snprintf(buffer, max, fmt, str);
  } else if (op == 'f') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if (file->filename != NULL)
      len = snprintf(buffer, max, fmt, file->filename);
  } else if (op == 'i') {
    snprintf(fmt, 512, "%%%sd", prefix);
    len = snprintf(buffer, max, fmt, index + 1);
  } else if (op == 'q') {
    snprintf(fmt, 512, "%%%sd", prefix);
    /* which number in queuelist? */
    if (file->inqueue)
      len = snprintf(buffer, max, fmt, getIDinList(queuelist, file) + 1);
  } else if (op == 'n') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((str = FUNC(getname, file, str_type_NAME)) != NULL)
      len = snprintf(buffer, max, fmt, str);
  } else if (op == 'a') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((str = FUNC(getname, file, str_type_ARTIST)) != NULL)
      len = snprintf(buffer, max, fmt, str);
  } else if (op == 'S') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((str = FUNC(getname, file, str_type_STATUS)) != NULL)
      len = snprintf(buffer, max, fmt, str);
  } else if (op == 's') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((time = FUNC(gettime, file)) > 0) {
      snprintf(dummy, 512, "%d:%02d", (time / 60), (time % 60));
      len = snprintf(buffer, max, fmt, dummy);
    }
  } else if (op == 'T') {
    snprintf(fmt, 512, "%%%ss", prefix);
    if ((total = FUNC(gettotal, file)) > 0) {
      snprintf(dummy, 512, "%d:%02d", (total / 60), (total % 60));
      len = snprintf(buffer, max, fmt, dummy);
    }
  }

  return len;
}

/*
 * parses a whole format entry
 * and parses all strings with parse_op
 */
int parseit(struct fmt_entry **fmt, char *buf, int bufsize, FILEENTRY file,
            int index) {
  struct fmt_entry **f;
  int len = 0;
  int ges = 0;
  char dummy[512];
  char prefix[512];
  char *cp;

  for (f = fmt; *f != NULL; f++) {
    if ((*f)->type == fmt_type_IFCLAUSE) {
      struct fmt_ifclause *ifc = (struct fmt_ifclause *) (*f);
      len = parseit(ifc->if_entry, dummy, bufsize, file, index);
      if (len > 0) {
        len = parseit(ifc->then_entry, buf, bufsize, file, index);
        ges += len;
        buf += len;
        bufsize -= len;
      } else {
        if (ifc->else_entry != NULL) {
          len = parseit(ifc->else_entry, buf, bufsize, file, index);
          ges += len;
          buf += len;
          bufsize -= len;
        }
      }
    } else if ((*f)->type == fmt_type_STRING) {
      char *ff = ((struct fmt_string *) (*f))->str;
      for (; *ff != 0; ff++) {
        if (*ff == '%') {
          ff++;
          cp = prefix;
          while (((*ff >= '0') && (*ff <= '9')) || *ff == '.' || *ff == '-')
            *cp++ = *ff++;
          *cp = 0;
          len = parse_op(*ff, prefix, buf, bufsize, file, index);
        } else {
          *buf = *ff;
          len = 1;
        }
        ges += len;
        buf += len;
        *buf = 0;
        bufsize -= len;
      }
      /* len = sprintf(buf, "%s", ((struct fmt_string *)(*f))->str); */
    } else if ((*f)->type == fmt_type_FILL) {
      struct fmt_filler *filler = (struct fmt_filler *) (*f);
      char dummy[512];
      len = parseit(filler->rest, dummy, 512, file, index);
      if (bufsize - len > 0)
        memset(buf, filler->filler, bufsize - len);
      buf += bufsize - len;
      ges += bufsize - len;
      memcpy(buf, dummy, len);
      buf += len;
      ges += len;
      *buf = 0;
      bufsize -= len;
    }
  }
  return ges;
}
