/* $Id: qplay.c 94 2005-03-12 16:43:09Z jesus $ */

#include <stdio.h>              /* stdio */
#include <stdlib.h>

#include <unistd.h>             /* getopt & .. */

#define __USE_GNU
#include <string.h>             /* str.. */

#include <ncurses.h>            /* ncurses */

#include <sys/socket.h>         /* socketpair */
#include <sys/timeb.h>          /* seektime */
#include <sys/time.h>
#include <sys/resource.h>       /* core dump */
#include <time.h>               /* time */

#define _GNU_SOURCE             /* getopt_long */
#include <getopt.h>             /* getopt_long */

#include <sys/types.h>          /* open */
#include <sys/stat.h>           /* open */
#include <fcntl.h>              /* open */

#include <signal.h>             /* signal */

#include <sys/wait.h>           /* wait */

/* include module functions */
#include "modules/mpg321.h"
#include "modules/ogg123.h"

#include "qplay.h"
#include "data.h"
#include "read-dir.h"
#include "read-m3u.h"
#include "actions.h"
#include "edit-list.h"
#include "globals.h"
#include "config.h"
#include "parser.h"

/*
 * draw the playlist window
 */
void redraw(FILELIST list, int startline) {
  int line;                     /* line on playlist window */
  // int time;                     /* current playing time in second */
  // int total;                    /* total playing time in seconds */

  if (rredraw) {                /* really redraw the screen, not only the changes */
    clear();
    redrawwin(stdscr);
    rredraw = 0;
  }

  /* print each line up to (LINES - 2) */
  for (line = 0; line < (LINES - 2); line++) {
    // clrtoeol();
    /* printing tracknumber and name */
    attrset(A_NORMAL);
    if (list->selected_file == (line + startline)) {  /* track is selected */
      attron(A_STANDOUT);       /* set selected color */
    }
    if (list->playing_file == (line + startline)) { /* track is playing_file */
      attron(COLOR_PAIR(CP_SELECTED));  /* set playing color scheme */
    }

    /* draw line by parsing format-string */
    move(line, 0);
    if ((line + startline) < list->count) {
      // printw(buf2, line + startline +1);
      char buffer[512];

      /* if (strcmp(format,"default") == 0)
         format = strdup("%i %{%a%t?%a # %t:%n}%> %q %{%s?%s/}%T"); */
      parseit(fmt, buffer, COLS, list->list[line + startline], line + startline);

      printw("%s", buffer);
    }
    clrtoeol();
  }

  /* clear the (LINES - 2)-line */
  move(line, 0);
  clrtoeol();

  /* set help line */
  move(LINES - 2, 0);
  attrset(COLOR_PAIR(CP_HELPLINE));
  /*
   * FIXME: create this line also by parsing a format string
   *        for now this is sufficient..
   */
  // addstr("\\ playing > ");
  int i;
#if 0
  printw("playing > ");
  for (i = 12; i < COLS - 40; i++)
    addch(' ');
  clrtoeol();
#endif

  if (list->count >= 1 && list->playing_file >= 0) {
    char buffer[512];

    // attron(COLOR_PAIR(CP_SELECTED));
    parseit(status_fmt, buffer, COLS - 48, list->list[list->playing_file], list->playing_file);
    printw("%s", buffer);

#if 0
    total = FUNC(gettotal, list->list[list->playing_file]);
    time = FUNC(gettime, list->list[list->playing_file]);
    parse_op('n', "", buf, 512, list->list[list->playing_file], list->playing_file);
    mvprintw(LINES - 2, 11, "%02d %s (%d:%02d/%d:%02d)", list->playing_file + 1,  /* list->list[list->playing_file]->filename, */
             buf, (time / 60), (time % 60), (total / 60), (total % 60));
#endif
  } else {
    for (i = 0; i < COLS; i++)
      addch(' ');
  }

  attrset(COLOR_PAIR(CP_HELPLINE));
  mvprintw(LINES - 2, COLS - 48,
           "S(O)rt  randomi(Z)e  (S)huffle [%c]  (R)epeat [%c]", shuffle ? 'X' : ' ',
           repeat ? 'X' : ' ');

  /* set input line */
  move(LINES - 1, 0);
  if (inputmode) {
    attrset(COLOR_PAIR(CP_INPUTLINE));
    printw(":");
    for (i = 0; i < inputchars; i++) {
      addch(inputline[i]);
    }
  } else if (errormode) {
    printw("%s", errorline);
  }

  clrtoeol();

  // doupdate();
#if 0
  if (rredraw) {
    redrawwin(stdscr);
    rredraw = 0;
  } else {
    refresh();
  }
#endif
  refresh();
}

/*
 * initializes default configuration
 */
void init_qplay() {

  /* format of the song list */
  // format = strdup("%i %{%a%t?%a # %t:%n} %> %q %{%s?%s/}%T");
  // char *default_fmt_string = "%i %{%a%t?%a # %t:%n} %> %q %{%s?%s/}%T";
  char *dummy = DEFAULT_FMT_STRING;
  fmt = str2fmt(&dummy);

  dummy = DEFAULT_STATUS_FMT_STRING;
  status_fmt = str2fmt(&dummy);

  /* initializes keymap */
  keymap = newKeymap();

  /*
   * adds default keybindings to keymap
   */
  add_keybinding(keymap, 0, '\n', "play");
  add_keybinding(keymap, 0, 's', "stop");

  add_keybinding(keymap, 0, 'd', "del");

  add_keybinding(keymap, 0, ' ', "pause");

  add_keybinding(keymap, CTRL_MODE, 'l', "clear");

  add_keybinding(keymap, 0, '/', "input search ");
  add_keybinding(keymap, CTRL_MODE, 'g', "search");

  add_keybinding(keymap, 0, 'n', "play next");
  add_keybinding(keymap, 0, 'p', "play prev");
  add_keybinding(keymap, 0, 'N', "play +5");
  add_keybinding(keymap, 0, 'P', "play -5");

  add_keybinding(keymap, 0, KEY_HOME, "select first");
  add_keybinding(keymap, 0, KEY_END, "select last");

  add_keybinding(keymap, 0, KEY_DOWN, "select next");
  add_keybinding(keymap, 0, KEY_UP, "select prev");
  add_keybinding(keymap, 0, KEY_NPAGE, "select +25");
  add_keybinding(keymap, 0, KEY_PPAGE, "select -25");
  add_keybinding(keymap, CTRL_MODE, 'f', "select +25");
  add_keybinding(keymap, CTRL_MODE, 'b', "select -25");
  add_keybinding(keymap, 0, 'j', "select next");
  add_keybinding(keymap, 0, 'k', "select prev");
  add_keybinding(keymap, 0, 'J', "move down");
  add_keybinding(keymap, 0, 'K', "move up");
  add_keybinding(keymap, 0, 'z', "select center");  //FIXME: should be key 'z' twice like in vi
  add_keybinding(keymap, CTRL_MODE, 'e', "select down");
  add_keybinding(keymap, CTRL_MODE, 'y', "select up");

  add_keybinding(keymap, 0, 'l', "seek +1");
  add_keybinding(keymap, 0, 'h', "seek -1");
  add_keybinding(keymap, 0, KEY_RIGHT, "seek +1");
  add_keybinding(keymap, 0, KEY_LEFT, "seek -1");

  add_keybinding(keymap, 0, KEY_IC, "insert2queue");
  add_keybinding(keymap, 0, KEY_DC, "deletefromqueue");

  add_keybinding(keymap, 0, ':', "input");

  add_keybinding(keymap, 0, 'Z', "randomize");
  add_keybinding(keymap, 0, 'O', "sort");

  add_keybinding(keymap, 0, 'R', "repeat");
  add_keybinding(keymap, 0, 'S', "shuffle");

  add_keybinding(keymap, 0, '?', "help");

  add_keybinding(keymap, 0, 'q', "quit");
}

/*
 * prints out user help screen
 */
void usage(char *program_name) {
  printf("Usage: %s [OPTION(s)] [FILE(s) | DIR(s)]\n", program_name);
  printf("\n"
         "supported options:\n"
         "  -h, --help                 print this help, then exit\n"
         "      --version              output version information and exit\n"
         "  -d, --debug[=LEVEL]        set debug output level\n"
         "  -f, --file[=FILE]          load config file\n"
         "\n"
         "  -a, --add-all              add all file formats\n"
         "                               (by default only supported file formats\n"
         "                               are added to the list)\n"
         "                               /* not jet implemented */\n"
         "  -r, --recursive            add subdirectories recursively\n"
         "  -i, --read-info            reads file information at startup\n"
         "                               (by default the information is read\n"
         "                               on play)\n"
         "                               /* not jet implemented */\n"
         "\n"
         "  -R, -l, --repeat           repeat playlist\n"
         "  -s, --shuffle              shuffle play\n"
         "  -o, --sort                 sort list\n"
         "  -Z, --randomize            randomize list\n"
         "  -L                         read playlist in m3u-format, as created by xmms\n"
         "\n" "  -u                         unlink fifo if it exists\n");
  exit(EXIT_FAILURE);
}

/* FIXME: fifo file */
#define FIFO_FILE "/tmp/qplay-fifo"

/*
 * handler for segfault signal
 */
void sign_segv(int signo) {
  if (signo == SIGSEGV) {
    if (filelist->playing_file >= 0)
      FUNC(stop, filelist->list[filelist->playing_file]);
    unlink(FIFO_FILE);
    signal(SIGSEGV, SIG_DFL);
  }
}

/*
 * handler for quit
 */
void bye() {
  endwin();
  unlink(FIFO_FILE);
  exit(0);
}

/*
 * handler for win-change signal
 * (not in use up to now)
 */
void sign_winch(int signo) {
  if (signo == SIGWINCH) {
    int i;
    for (i = 0; i < 10; i++) {
      printf("SIGWINCH %d %d\r\n", COLS, LINES);
    }
    fflush(stdout);
    getmaxyx(stdscr, LINES, COLS);
  }
}

int main(int argc, char *argv[]) {
  fd_set fds;                   /* set of file descriptor for select */
  struct timeval timeout;       /* timeout for select */

  /* struct timeb seeknow;
     struct timeb seeklast;
     int seekvalue; */
  int key;                      /* received key */
  int i;                        /* index */
  // int new_remote;
  char buf[512];                /* buffer for reading from player and
                                   remote fifo */
  int recursive = 0;            /* directories-search recursively */
  int loadm3ulist = 0;          /* read playlist in m3u-format ? */

  struct pathlist plist;
  // char **pathlist;              /* list of path to scan - FIXME */
  int randomize = 0;            /* randomize status */
  int sort = 0;                 /* sort status */

  char *configfile = NULL;

  /* creates a new FileList for the queue list */
  queuelist = newFileList();

  int o;                        /* small option */
  int option_index;             /* index of long option */

  int unlink_fifo = 0;

  /* module functions */
  CHECK_HANDLER checks[] = { mpg321_check, ogg123_check };
  INIT_HANDLER inits[] = { mpg321_init, ogg123_init };
  int modules =                 /* number of modules */
    sizeof(checks) / sizeof(CHECK_HANDLER);

  /* long options for getopt_long */
  static struct option long_options[] = {
    {"help", no_argument, NULL, 'h'},
    {"version", no_argument, NULL, 0},
    {"debug", optional_argument, NULL, 'd'},

    {"command", required_argument, NULL, 'c'},

    {"add-all", no_argument, NULL, 'a'},
    {"recursive", no_argument, NULL, 'r'},
    {"read-info", no_argument, NULL, 'i'},

    {"shuffle", no_argument, NULL, 's'},
    {"repeat", no_argument, NULL, 'l'}, /* R */
    {"randomize", no_argument, NULL, 'Z'},
    {"sort", no_argument, NULL, 'o'},

    {"file", required_argument, NULL, 'f'},
    {"load", required_argument, NULL, 'f'},

    {0, 0, 0, 0}
  };

  /*
   * first of all we parse the parameter line
   */
  while ((o =
          getopt_long(argc, argv, "d::rhosZRailc:f:Lu", long_options,
                      &option_index)) >= 0) {
    switch (o) {
    case 0:                    /* long option */
      if (strcmp(long_options[option_index].name, "version") == 0) {
        printf(PACKAGE " " VERSION "\n");
        printf("Written by jesus c. meyendriesch and Alexander Fieroch.\n\n"
               "Copyright (C) 2004 Free Software Foundation, Inc.\n"
               "This is free software; see the source for copying conditions.  There is NO\n"
               "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
        exit(EXIT_SUCCESS);
      }
      break;

    case 'c':                  /* command */
      {
        fd_set fds;
        int fifo;
        struct stat filestat;

        if ((stat(FIFO_FILE, &filestat) == 0) && (S_ISFIFO(filestat.st_mode) > 0)) {
          if ((fifo = open("/tmp/qplay-fifo", O_WRONLY)) > 0) {
            FD_ZERO(&fds);
            FD_SET(fifo, &fds);
            if (select(8, NULL, &fds, NULL, NULL) > 0)
              if (FD_ISSET(fifo, &fds))
                write(fifo, optarg, strlen(optarg));
            close(fifo);
          }
        }
      }

      exit(EXIT_SUCCESS);
      break;

    case 'f':                  /* load file */
      configfile = strdup(optarg);
      break;

    case 'a':                  /* add-all */
      break;

    case 'i':                  /* read-info */
      break;

    case 'l':                  /* repeat */
    case 'R':
      repeat = 1;
      break;

    case 's':                  /* shuffle */
      shuffle = 1;
      break;

    case 'o':                  /* sort */
      sort = 1;
      break;

    case 'Z':                  /* randomize */
      randomize = 1;
      break;

    case 'r':                  /* recursive */
      recursive = 1;
      break;

    case 'L':                  /* load m3u-list */
      loadm3ulist = 1;
      break;

    case 'h':                  /* help */
      usage(argv[0]);
      break;

    case 'u':                  /* unlink */
      unlink_fifo = 1;
      break;

    case 'd':                  /* debug */
      if (optarg)
        debug = atoi(optarg);
      else
        debug = 1;
      break;

    case '?':
      usage(argv[0]);
      break;

    default:
      printf("?? getopt returned character code 0x%02x ??\n", o);
      exit(EXIT_FAILURE);
      break;
    }
  }

  /*
   * look for left parameters
   */
  if (optind < argc) {
    /* create a list of the rest parameters */
    plist.count = (argc - optind);
    plist.max = plist.count;
    plist.path = malloc(sizeof(char *) * plist.count);
    for (i = 0; optind < argc; i++) {
      plist.path[i] = strdup(argv[optind++]);
    }
  } else {
    /* if no other options are given, create a list
     * with only one path (".") */
    if (loadm3ulist) {
      fprintf(stderr, "%s: option -L is given, but no playlist files are specified!",
              argv[0]);
      exit(EXIT_FAILURE);
    }

    plist.count = 0;
    plist.max = 1;
    plist.path = malloc(sizeof(char *));

    /*
    plist.count = 1;
    plist.max = 1;
    plist.path = malloc(sizeof(char *));
    plist.path[0] = strdup(".");
    */
  }

  /*
   * initializes some global variables and key settings
   */
  init_qplay();

  /*
   * reads direcotries from pathlist
   * and create filelist for all files
   */
  if (loadm3ulist) {            /* FIXME !! THIS IS ONLY TEMPORARY !! */
    filelist = get_file_list_from_m3u(&plist);
    int i, j;
    for (i = 0; i < filelist->count;) {
      /* checks all modules, in order to handle the file */
      for (j = 0; j < modules; j++) {
        if (checks[j] (filelist->list[i])) {
          i++;
          break;
        }
      }
      /* if no module were found, the file is removed from the list */
      /* FIXME: there should also be a free here */
      if (j == modules) {
        if (filelist->count > 1) {
          for (j = i; j < (filelist->count - 1); j++) {
            filelist->list[j] = filelist->list[j + 1];
          }
        }
        filelist->count--;
      }
    }
    //    filelist = parseFile(pathlist[0]);
  } else {
    /* creates only the filelist and reads the files later
     * in its own fork */
    filelist = newFileList();
  }

  /*
   * if randomize is given at command line
   */
  if (randomize) {
    if ((errorline = command("randomize")) != NULL) {
      errormode = 1;
    } else {
      errormode = 0;
    }
  }

  /*
   * if sort is given at command line use command
   * system to sort
   */
  if (sort) {
    if ((errorline = command("sort")) != NULL) {
      errormode = 1;
    } else {
      errormode = 0;
    }
  }

  /*
   * initializes all modules
   *   calls init-functions of each module
   */
  for (i = 0; i < modules; i++) {
    inits[i] (argc, argv);
  }

  /*
   * initializes curses
   */
  initscr();

  /* FIXME */
  start_color();

  init_pair(CP_DEFAULT, COLOR_GREEN, COLOR_BLACK);
  init_pair(CP_SELECTED, COLOR_GREEN, COLOR_BLACK);
  init_pair(CP_HELPLINE, COLOR_RED, COLOR_WHITE);
  init_pair(CP_INPUTLINE, COLOR_WHITE, COLOR_BLACK);

  raw();                        /* FIXME */

  cbreak();
  noecho();
  nodelay(stdscr, TRUE);        /* never waits for a key *
                                 * select will do it for us */

  keypad(stdscr, TRUE);
  // meta(stdscr, TRUE);    /* ??? */

  /*
   * prepare both set of filedescriptor for select
   *   sets to zero
   */
  FD_ZERO(&fds);                /* prepare filedescription selection */
  FD_ZERO(&fds_copy);

  struct stat statbuf;

  /*
   * if no config file is given as command line parameter,
   * we try to use the default config file ~/.qplayrc
   */
  if (configfile == NULL) {
    char *home;
    char path[512];

    home = getenv("HOME");
    sprintf(path, "%s/.qplayrc", home);
    if (stat(path, &statbuf) == 0) {
      configfile = path;
    }
  }

  /*
   * now let us read all config-file lines and give them
   * to command as parameter
   */
  FILE *infile;
  if ((configfile != NULL) &&
      (stat(configfile, &statbuf) == 0) &&
      ((infile = fopen(configfile, "r")) != NULL)) {
    while (fgets(buf, 512, infile) != NULL) {
      char *ch;
      if (buf[0] == '#')        /* ignore comments */
        continue;

      for (ch = buf; *ch != 0; ch++) {
        if (*ch == '\n') {
          *ch = 0;
          break;
        }
      }
      if ((errorline = command(buf)) != NULL) {
        /* print out errors */
        errormode = 1;
        printw("error in config line: '%s'\n", buf);
        printw("%s\n", errorline);
        printw("press <esc> to exit or any key to continue");
        if (getch() == '\033') {
          endwin();
          exit(1);
        }
      } else {
        errormode = 0;
      }
    }
    fclose(infile);
  }

  /*
   * sets core file size to the max
   * therefore a file with memory dump is caused on
   * a segv-signal
   */
  if (debug > 0) {
    struct rlimit rlim;
    getrlimit(RLIMIT_CORE, &rlim);
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_CORE, &rlim);
  }

  signal(SIGSEGV, sign_segv);   /* sets segv-handler to stop player on segfault */
  signal(SIGQUIT, bye);         /* sets quit-handler to unlink fifo */
  signal(SIGINT, bye);          /* sets quit-handler to unlink fifo */
  // signal(SIGWINCH, sign_winch);  /* */
  // atexit(bye);

  /*
   * removes fifo if it exists and unlink_fifo is set
   */
  if (unlink_fifo && stat(FIFO_FILE, &statbuf) == 0) {
    unlink(FIFO_FILE);
  }

  /*
   * creates new fifo file
   */
  int fifo, com[2];
  pid_t pid;

  if (mknod(FIFO_FILE, S_IFIFO | 0644, 0) != 0) {
    endwin();
    ERROR(1, "qplay already running. Please kill process or remove " FIFO_FILE "\n");
  }

  fifo = open(FIFO_FILE, O_RDONLY | O_NDELAY);
  FD_SET(fifo, &fds_copy);

  /* FIXME we have to intercept errors */
  pipe(com);

  /*
   * creates a child in order to read in the files
   * from command-line
   */
  if ((pid = fork() == 0)) {    /* child */
    close(com[0]);
    get_file_list(com[1], &plist, recursive);
    exit(0);
  } else if (pid < 0) {         /* error */
    /* FIXME */
  } /* mother */
  close(com[1]);
  FD_SET(com[0], &fds_copy);

  /*
   * adds stdin to a set of file descriptors to
   * read keypress
   */
  FD_SET(0, &fds_copy);

  /*
   * at first draw screen
   */
  redraw(filelist, startline);

  /* forever */
  for (;;) {
    timeout.tv_sec = 10;        /* set timeout for select this is */
    timeout.tv_usec = 0;        /* not important. we can try it later
                                 * without that */
    fds = fds_copy;             /* select modifies fds, so there is a copy
                                 * and we have to change this to make
                                 * selection */

    PRINT_DEBUG(2, "startline: %d - playing_file: %d - selected_file: %d\n",
                startline, filelist->playing_file, filelist->selected_file);
    /*
     * waits for data of one of the filedescriptors in
     * the set
     */
    if (select(8, &fds, NULL, NULL, &timeout) > 0) {
      /* keypress arrived */
      if (FD_ISSET(0, &fds)) {
        int meta_key = 0;
        key = getch();          /* read it from ncurses */
        if (key == '\033') {    /* maybe a meta key */
          meta_key = 1;
          // nodelay(stdscr, 1);   /* don't wait for key */
          if ((key = getch()) == ERR) {
            key = '\033';
            meta_key = 0;
          }
          // nodelay(stdscr, 0);
        }
        PRINT_DEBUG(2, "Key: %d %d\n", meta_key, key);
        if (inputmode) {        /* if there is the input mode */
          switch (key) {
          case KEY_BACKSPACE:
            if (inputchars > 1) {
              inputchars--;
              inputline = realloc(inputline, sizeof(char) * inputchars);
            } else if (inputchars > 0) {
              inputchars = 0;
              free(inputline);
            }
            break;
          case '\n':
            inputmode = 0;
            if (inputchars > 0) {
              char *dummy;      /* FIXED problem with "input foo" command */
              inputline = realloc(inputline, sizeof(char) * (inputchars + 1));
              inputline[inputchars] = 0;
              dummy = inputline;
              errorline = command(inputline);
              if (errorline != NULL) {
                errormode = 1;
              } else {
                errormode = 0;
              }
              free(dummy);
              if (dummy == inputline)
                inputchars = 0;
            }
            break;
          default:
            if (inputchars) {
              inputline = realloc(inputline, sizeof(char) * (inputchars + 1));
            } else {
              inputline = malloc(sizeof(char));
            }
            inputline[inputchars++] = key;
            break;
          }
        } else {
          /*
           * looks for key in keymap
           */
          for (i = 0; i < keymap->count; i++) {
            if (((keymap->list[i].keymode & META_MODE) && meta_key)
                || (!(keymap->list[i].keymode & META_MODE) && !meta_key)) {
              if (keymap->list[i].keycode == key) {
                errorline = command(keymap->list[i].action);
                if (errorline != NULL) {
                  errormode = 1;
                } else {
                  errormode = 0;
                }
                break;
              }
            }
          }
          /*
           * if key was not found
           */
          if (i == keymap->count) {
            if (debug > 0) {
              printf("unbound key: 0x%02x", key);
              if (key > 0x2f) {
                printf(" (%c)", key);
              }
              printf("\r\n");
            }
          }
#if 0                           /* {{{ old keymap system */
          if (i == keymap->count) {
            switch (key) {

#define PLAY { \
  if (filelist->playing_file >= 0) { \
    FUNC(play, filelist->list[filelist->playing_file]); \
    remote = FUNC(getfd, filelist->list[filelist->playing_file]); \
    FD_SET(remote, &fds_copy); \
  } \
}

#define STOP { \
  if (filelist->playing_file >= 0) { \
    FUNC(stop, filelist->list[filelist->playing_file]); \
    FD_CLR(remote, &fds_copy); \
  } \
}

            case 'S':          /* Shuffle */
              if (shuffle)
                shuffle = 0;
              else
                shuffle = 1;
              break;
            case 'R':          /* Repeat */
              if (repeat)
                repeat = 0;
              else
                repeat = 1;
              break;

            case 'p':          /* previous */
              if (shuffle) {    /* FIXME: in shuffle mode next and previous both have the same function */
                STOP;
                filelist->playing_file =
                  (int) (filelist->count * (rand() / (RAND_MAX + 1.0)));
                PLAY;
              } else if (filelist->playing_file > 0) {
                FUNC(stop, filelist->list[filelist->playing_file]);
                FD_CLR(remote, &fds_copy);

                filelist->playing_file--;
                PLAY;
              } else if (repeat) {
                STOP;
                filelist->playing_file = filelist->count - 1;
                PLAY;
              }
              break;
            case 'n':          /* next */
              if (queuelist != NULL && queuelist->count > 0) {
                STOP;
//            playing_file = queuelist->list[0];
                int j;
                for (j = 0; j < filelist->count; j++) {
                  if (filelist->list[j] == queuelist->list[0])
                    filelist->playing_file = j;
                }
                filetoplay = queuelist->list[0];
                removeFromList(queuelist, queuelist->list[0]);
                //PLAY;
                FUNC(play, filetoplay);
                remote = FUNC(getfd, filetoplay);
                fprintf(stderr, "\n\nqueueplaying: %s", filetoplay->filename);
                FD_SET(remote, &fds_copy);
              } else if (shuffle) { /* FIXME: in shuffle mode next and previous both have the same function */
                STOP;
                filelist->playing_file =
                  (int) (filelist->count * (rand() / (RAND_MAX + 1.0)));
                PLAY;
              } else if (((filelist->playing_file + 1) < filelist->count)
                         && (filelist->playing_file >= 0)) {
                FUNC(stop, filelist->list[filelist->playing_file]);
                FD_CLR(remote, &fds_copy);

                filelist->playing_file++;
                PLAY;
              } else if (repeat) {
                STOP;
                filelist->playing_file = 0;
                PLAY;
              }
              break;
            case 's':          /* stop */
              STOP;
              filelist->playing_file = NONE;
              break;
            case ' ':          /* pause */
              if (filelist->playing_file >= 0)
                FUNC(pause, filelist->list[filelist->playing_file]);
              if (remote > 0)
                FD_CLR(remote, &fds_copy);
              if ((new_remote =
                   FUNC(getfd, filelist->list[filelist->playing_file])) > 0) {
                remote = new_remote;
                FD_SET(remote, &fds_copy);
              }
              break;
            case KEY_IC:       /* inserts selected_file to queuelist */
//              addToList(queuelist, filelist->list[filelist->selected_file]);
              //FIXME DOWN(1);

              /* DEBUG */
              if (debug > 1) {
                int i;
                fprintf(stderr, "\nINSERT: %i, count: %i, max: %i",
                        filelist->selected_file, queuelist->count, queuelist->max);
                for (i = 0; i < queuelist->count; i++)
                  fprintf(stderr, "\nfile in queue: %s",
                          queuelist->list[i]->filename);
                fprintf(stderr, "\n");
              }
              break;
            case KEY_DC:       /* delets selected_file from queuelist */
              removeFromList(queuelist, filelist->list[filelist->selected_file]);
              //FIXME DOWN(1);

              /* DEBUG */
              if (debug > 1) {
                fprintf(stderr, "\nDELETE:   count: %i, max: %i", queuelist->count,
                        queuelist->max);
                for (i = 0; i < queuelist->count; i++)
                  fprintf(stderr, "\nfile in queue: %s",
                          queuelist->list[i]->filename);
                fprintf(stderr, "\n");
              }
              break;
            default:
              PRINT_DEBUG(2, "unknown key: %d", key);
              break;
            } /* switch (key) */
          }
#endif /* }}} */
        }


      } else if (FD_ISSET(remote, &fds)) {
        /*
         * remote player process has output
         */
        int len;

        if ((len = read(remote, buf, 512)) > 0) {
          if (len < 512)
            buf[len] = 0;
          if (debug > 2)
            write(1, buf, len);
          FUNC(parse, filelist->list[filelist->playing_file], buf);
        } else {                /* player ends */
          command("play next");
        }
      } else if (FD_ISSET(fifo, &fds)) {
        int len;

        if ((len = read(fifo, buf, 512)) > 0) {
          buf[len] = 0;
          // write(1, buf, len);
          command(buf);
          // sleep(3);
        } else {
          if (fifo > 0) {
            close(fifo);
            FD_CLR(fifo, &fds_copy);
            fifo = open(FIFO_FILE, O_RDONLY | O_NDELAY);
            FD_SET(fifo, &fds_copy);
          }
        }
      } else if (FD_ISSET(com[0], &fds)) {
        /*
         * remote process found a new file
         */
        int len, size;

        /* FIXME FIXME FIXME FIXME FIXME */
        if (read(com[0], &size, sizeof(int)) > 0) {
          if ((len = read(com[0], buf, size)) > 0) {
            FILEENTRY new;
            char *filename;
            int j;

            buf[len] = 0;

            filename = strdup(buf);
            new = newFileEntry(filename);
            for (j = 0; j < modules; j++) {
              if (checks[j] (new)) {
                addToList(filelist, new);
                break;
              }
            }
          }
        } else {
          FD_CLR(com[0], &fds_copy);
          close(com[0]);
          wait(NULL);
          if (randomize) {
            command("randomize");
            randomize = 0;
          }
        }
      }
      redraw(filelist, startline);  /* redraw screen every loop */
    } /* if (select(8, &fds, NULL, NULL, &timeout) > 0) */
  } /* for (;;) */

  FUNC(stop, filelist->list[filelist->playing_file]);
  close(remote);

  endwin();

  return 0;
}

// vim:tw=78:ts=2:ft=c
// vim600:sw=2:sts=2:fdm=marker:si:sta:et
